# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Building 'Segment' Level Data Attributes"""

from copy import copy

try:
    from typing import Any, Dict, Union
except ImportError:
    pass  # Python2.7

from honeybee_energy_ph.construction import thermal_bridge
from honeybee_ph import _base, phi, phius, site
from honeybee_ph_standards.sourcefactors import factors
from honeybee_ph_utils import enumerables


class PhVentilationSummerBypassMode(enumerables.CustomEnum):
    allowed = ["1-None", "2-Temperatur Controlled", "3-Enthalpy Controlled", "4-Always"]

    def __init__(self, _value=1):
        # type: (Union[str, int]) -> None
        super(PhVentilationSummerBypassMode, self).__init__(_value)


class SetPoints(_base._Base):
    def __init__(self):
        super(SetPoints, self).__init__()
        self.winter = 20.0
        self.summer = 25.0

    def to_dict(self):
        # type: () -> Dict[str, float]
        d = {}

        d["winter"] = self.winter
        d["summer"] = self.summer
        d["user_data"] = copy(self.user_data)

        return d

    @classmethod
    def from_dict(cls, _dict):
        # type: (Dict[str, Any]) -> SetPoints
        obj = cls()

        obj.winter = _dict.get("winter", 20.0)
        obj.summer = _dict.get("summer", 25.0)
        obj.user_data = _dict.get("user_data", {})

        return obj

    def __copy__(self):
        # type: () -> SetPoints
        obj = SetPoints()
        obj.set_base_attrs_from_source(self)
        obj.winter = self.winter
        obj.summer = self.summer
        obj.user_data = self.user_data

        return obj

    def duplicate(self):
        # type: () -> SetPoints
        return self.__copy__()


class BldgSegment(_base._Base):
    def __init__(self):
        super(BldgSegment, self).__init__()
        self.display_name = "Unnamed_Bldg_Segment"
        self.num_floor_levels = 1
        self.num_dwelling_units = 1
        self.site = site.Site()
        self.source_energy_factors = factors.FactorCollection("Source_Energy")
        self.co2e_factors = factors.FactorCollection("CO2")
        self.phius_certification = phius.PhiusCertification()
        self.phi_certification = phi.PhiCertification()
        self.set_points = SetPoints()
        self.mech_room_temp = 20.0
        self.non_combustible_materials = False
        self.thermal_bridges = {}  # type: Dict[str, thermal_bridge.PhThermalBridge]
        self.summer_hrv_bypass_mode = PhVentilationSummerBypassMode("4-Always")

    def add_new_thermal_bridge(self, tb):
        # type: (thermal_bridge.PhThermalBridge) -> None
        self.thermal_bridges[tb.identifier] = tb

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = {}

        d["identifier"] = self.identifier
        d["name"] = self.display_name
        d["num_floor_levels"] = self.num_floor_levels
        d["num_dwelling_units"] = self.num_dwelling_units
        d["site"] = self.site.to_dict()
        d["source_energy_factors"] = self.source_energy_factors.to_dict()
        d["co2e_factors"] = self.co2e_factors.to_dict()
        d["phius_certification"] = self.phius_certification.to_dict()
        d["phi_certification"] = self.phi_certification.to_dict()
        d["set_points"] = self.set_points.to_dict()
        d["mech_room_temp"] = self.mech_room_temp
        d["non_combustible_materials"] = self.non_combustible_materials
        d["thermal_bridges"] = {}
        for tb in self.thermal_bridges.values():
            d["thermal_bridges"][str(tb.identifier)] = tb.to_dict()
        d["user_data"] = self.user_data
        d["summer_hrv_bypass_mode"] = self.summer_hrv_bypass_mode.to_dict()
        return d

    @classmethod
    def from_dict(cls, _dict):
        # type: (Dict[str, Any]) -> BldgSegment
        obj = cls()

        obj.identifier = _dict.get("identifier")
        obj.display_name = _dict.get("name")
        obj.num_floor_levels = _dict.get("num_floor_levels")
        obj.num_dwelling_units = _dict.get("num_dwelling_units")
        obj.site = site.Site.from_dict(_dict.get("site", {}))
        obj.source_energy_factors = factors.FactorCollection.from_dict(_dict.get("source_energy_factors", {}))
        obj.co2e_factors = factors.FactorCollection.from_dict(_dict.get("co2e_factors", {}))
        obj.phius_certification = phius.PhiusCertification.from_dict(_dict.get("phius_certification", {}))
        obj.phi_certification = phi.PhiCertification.from_dict(_dict.get("phi_certification", {}))
        obj.set_points = SetPoints.from_dict(_dict.get("set_points", {}))
        obj.mech_room_temp = _dict["mech_room_temp"]
        obj.non_combustible_materials = _dict.get("non_combustible_materials", False)
        for tb_dict in _dict["thermal_bridges"].values():
            tb_obj = thermal_bridge.PhThermalBridge.from_dict(tb_dict)
            obj.thermal_bridges[tb_obj.identifier] = tb_obj
        obj.user_data = _dict.get("user_data", {})
        obj.summer_hrv_bypass_mode = PhVentilationSummerBypassMode.from_dict(_dict.get("summer_hrv_bypass_mode", {}))
        return obj

    def __copy__(self):
        # type () -> BldgSegment
        new_obj = BldgSegment()

        new_obj.identifier = self.identifier
        new_obj.display_name = self.display_name
        new_obj.num_floor_levels = self.num_floor_levels
        new_obj.num_dwelling_units = self.num_dwelling_units
        new_obj.site = self.site.duplicate()
        new_obj.source_energy_factors = self.source_energy_factors.duplicate()
        new_obj.co2e_factors = self.co2e_factors.duplicate()
        new_obj.phius_certification = self.phius_certification.duplicate()
        new_obj.phi_certification = self.phi_certification.duplicate()
        new_obj.set_points = self.set_points.duplicate()
        new_obj.mech_room_temp = self.mech_room_temp
        new_obj.non_combustible_materials = self.non_combustible_materials
        new_obj.thermal_bridges = {}
        for tb_k, tb_v in self.thermal_bridges.items():
            new_obj.thermal_bridges[tb_k] = tb_v.duplicate()
        new_obj.user_data = self.user_data
        new_obj.summer_hrv_bypass_mode = PhVentilationSummerBypassMode(self.summer_hrv_bypass_mode.value)
        return new_obj

    def duplicate(self):
        # type () -> BldgSegment
        return self.__copy__()



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Honeybee-PH-HVAC-Equipment: Ventilation (ERV) Devices."""

import sys
from copy import copy

try:
    from typing import Any, Dict, List, Optional
except ImportError:
    pass  # IronPython

try:
    from ladybug_geometry.geometry3d.pointvector import Point3D
except ImportError as e:
    raise ImportError("Failed to import ladybug_geometry", e)

try:
    from honeybee_phhvac import _base, ducting
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_phhvac:\n\t{}".format(e))


# -----------------------------------------------------------------------------


class UnknownPhExhaustVentTypeError(Exception):
    def __init__(self, _device_types, _received_type):
        # type: (list[str], str) -> None
        self.msg = 'Error: Unknown HBPH-Exhaust Ventilation type? Got: "{}"\
            "but only types: {} are allowed?'.format(
            _received_type, _device_types
        )
        super(UnknownPhExhaustVentTypeError, self).__init__(self.msg)


# -----------------------------------------------------------------------------
# -- ERV units


class Ventilator(_base._PhHVACBase):
    def __init__(self):
        super(Ventilator, self).__init__()
        self.display_name = "_unnamed_ventilator_"  # type: str
        self.id_num = 0  # type: int
        self.quantity = 1  # type: int
        self.sensible_heat_recovery = 0.0  # type:float
        self.latent_heat_recovery = 0.0  # type float
        self.electric_efficiency = 0.55  # type: float
        self.frost_protection_reqd = True  # type: bool
        self.temperature_below_defrost_used = -5  # type: float
        self.in_conditioned_space = True  # type: bool

    def to_dict(self):
        # type: () -> dict[str, Any]
        d = super(Ventilator, self).to_dict()
        d["quantity"] = self.quantity
        d["sensible_heat_recovery"] = self.sensible_heat_recovery
        d["latent_heat_recovery"] = self.latent_heat_recovery
        d["electric_efficiency"] = self.electric_efficiency
        d["frost_protection_reqd"] = self.frost_protection_reqd
        d["temperature_below_defrost_used"] = self.temperature_below_defrost_used
        d["in_conditioned_space"] = self.in_conditioned_space
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict[str, Any]) -> Ventilator
        obj = cls()
        obj.display_name = _input_dict["display_name"]
        obj.identifier = _input_dict["identifier"]
        obj.user_data = _input_dict.get("user_data", {})
        obj.quantity = _input_dict["quantity"]
        obj.sensible_heat_recovery = _input_dict["sensible_heat_recovery"]
        obj.latent_heat_recovery = _input_dict["latent_heat_recovery"]
        obj.electric_efficiency = _input_dict["electric_efficiency"]
        obj.frost_protection_reqd = _input_dict["frost_protection_reqd"]
        obj.temperature_below_defrost_used = _input_dict["temperature_below_defrost_used"]
        obj.in_conditioned_space = _input_dict["in_conditioned_space"]
        return obj

    def duplicate(self):
        # type: () -> Ventilator
        new_obj = self.__class__()
        new_obj.display_name = self.display_name
        new_obj.identifier = self.identifier
        new_obj.user_data = copy(self.user_data)
        new_obj.id_num = self.id_num
        new_obj.quantity = self.quantity
        new_obj.sensible_heat_recovery = self.sensible_heat_recovery
        new_obj.latent_heat_recovery = self.latent_heat_recovery
        new_obj.electric_efficiency = self.electric_efficiency
        new_obj.frost_protection_reqd = self.frost_protection_reqd
        new_obj.temperature_below_defrost_used = self.temperature_below_defrost_used
        new_obj.in_conditioned_space = self.in_conditioned_space
        return new_obj

    def __lt__(self, other):
        # type: (PhVentilationSystem) -> bool
        return self.identifier < other.identifier

    def __copy__(self):
        # type: () -> Ventilator
        return self.duplicate()

    def __repr__(self):
        return "{}(display_name={!r}, sensible_heat_recovery={:0.2f})".format(
            self.__class__.__name__, self.display_name, self.sensible_heat_recovery
        )

    def ToString(self):
        return self.__repr__()


class PhVentilationSystem(_base._PhHVACBase):
    """Passive House Fresh-Air Ventilation System."""

    def __init__(self):
        # type: () -> None
        super(PhVentilationSystem, self).__init__()
        self.display_name = "_unnamed_ph_vent_system_"
        self.sys_type = 1  # '1-Balanced PH ventilation with HR'
        self.supply_ducting = []  # type: List[ducting.PhDuctElement]
        self.exhaust_ducting = []  # type: List[ducting.PhDuctElement]
        self._ventilation_unit = None  # type: Optional[Ventilator]
        self.id_num = 0

    @property
    def ventilation_unit(self):
        # type: () -> Optional[Ventilator]
        return self._ventilation_unit

    @ventilation_unit.setter
    def ventilation_unit(self, _in):
        # type: (Optional[Ventilator]) -> None
        self._ventilation_unit = _in

        if not self._ventilation_unit:
            return None

        if self._ventilation_unit.display_name == "_unnamed_ventilator_":
            self._ventilation_unit.display_name = self.display_name
        return None

    def add_supply_duct_element(self, _duct_element):
        # type: (ducting.PhDuctElement) -> None
        """Add a supply-air duct element to the ventilation system."""
        self.supply_ducting.append(_duct_element)

    def add_exhaust_duct_element(self, _duct_element):
        # type: (ducting.PhDuctElement) -> None
        """Add an exhaust-air duct element to the ventilation system."""
        self.exhaust_ducting.append(_duct_element)

    @property
    def supply_ducting_total_length(self):
        # type: () -> float
        """Return the total length of all supply-air ducting in model-units."""
        return sum(duct.length for duct in self.supply_ducting)

    @property
    def exhaust_ducting_total_length(self):
        # type: () -> float
        """Return the total length of all exhaust-air ducting in model-units."""
        return sum(duct.length for duct in self.exhaust_ducting)

    @property
    def supply_ducting_size_description(self):
        # type: () -> Optional[str]
        """Return the size of the supply-air ducting."""
        descriptions = {s.shape_type_description for s in self.supply_ducting}
        if len(descriptions) == 0:
            return None
        elif len(descriptions) == 1:
            return descriptions.pop()
        else:
            raise ValueError("Mixed shape-types in supply-air duct segments.")

    @property
    def exhaust_ducting_size_description(self):
        # type: () -> Optional[str]
        """Return the size of the exhaust-air ducting."""
        descriptions = {s.shape_type_description for s in self.exhaust_ducting}
        if len(descriptions) == 0:
            return None
        elif len(descriptions) == 1:
            return descriptions.pop()
        else:
            raise ValueError("Mixed shape-types in exhaust-air duct segments.")

    def to_dict(self):
        # type: () -> dict[str, Any]
        d = super(PhVentilationSystem, self).to_dict()
        d["sys_type"] = self.sys_type
        d["exhaust_ducting"] = [e_duct.to_dict() for e_duct in self.exhaust_ducting]
        d["supply_ducting"] = [s_duct.to_dict() for s_duct in self.supply_ducting]
        d["id_num"] = self.id_num
        if self.ventilation_unit:
            d["ventilation_unit"] = self.ventilation_unit.to_dict()
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict[str, Any]) -> PhVentilationSystem
        obj = cls()

        obj.identifier = _input_dict["identifier"]
        obj.display_name = _input_dict["display_name"]
        obj.user_data = _input_dict.get("user_data", {})
        obj.sys_type = _input_dict["sys_type"]
        obj.supply_ducting = [ducting.PhDuctElement.from_dict(s_duct) for s_duct in _input_dict["supply_ducting"]]
        obj.exhaust_ducting = [ducting.PhDuctElement.from_dict(e_duct) for e_duct in _input_dict["exhaust_ducting"]]
        obj.id_num = _input_dict.get("id_num", 0)

        vent_unit_dict = _input_dict.get("ventilation_unit", None)
        if vent_unit_dict:
            obj.ventilation_unit = Ventilator.from_dict(vent_unit_dict)

        return obj

    def duplicate(self):
        # type: () -> PhVentilationSystem
        new_obj = self.__class__()
        new_obj.display_name = self.display_name
        new_obj.identifier = self.identifier
        new_obj.user_data = copy(self.user_data)
        new_obj.sys_type = self.sys_type
        new_obj.supply_ducting = [s_duct.duplicate() for s_duct in self.supply_ducting]
        new_obj.exhaust_ducting = [e_duct.duplicate() for e_duct in self.exhaust_ducting]
        new_obj.id_num = self.id_num

        if self.ventilation_unit:
            new_obj._ventilation_unit = self.ventilation_unit.duplicate()

        return new_obj

    def __copy__(self):
        # type: () -> PhVentilationSystem
        return self.duplicate()

    def __lt__(self, other):
        # type: (PhVentilationSystem) -> bool
        return self.identifier < other.identifier

    def __repr__(self):
        return "{}(display_name={!r}, sys_type={!r})".format(self.__class__.__name__, self.display_name, self.sys_type)

    def ToString(self):
        return self.__repr__()

    def move(self, moving_vec3D):
        """Move the System's ducts along a vector.

        Args:
            moving_vec3D: A Vector3D with the direction and distance to move the ray.
        """
        new_system = self.duplicate()
        new_system.identifier = self.identifier
        new_system.supply_ducting = [duct_element.move(moving_vec3D) for duct_element in self.supply_ducting]
        new_system.exhaust_ducting = [duct_element.move(moving_vec3D) for duct_element in self.exhaust_ducting]
        return new_system

    def rotate(self, axis_vec3D, angle_degrees, origin_pt3D):
        """Rotate the System's ducts by a certain angle around an axis and origin.

        Right hand rule applies:
        If axis has a positive orientation, rotation will be clockwise.
        If axis has a negative orientation, rotation will be counterclockwise.

        Args:
            axis_vec3D: A Vector3D axis representing the axis of rotation.
            angle_degrees: An angle for rotation in degrees.
            origin_pt3D: A Point3D for the origin around which the object will be rotated.
        """
        new_system = self.duplicate()
        new_system.identifier = self.identifier
        new_system.supply_ducting = [
            duct_element.rotate(axis_vec3D, angle_degrees, origin_pt3D) for duct_element in self.supply_ducting
        ]
        new_system.exhaust_ducting = [
            duct_element.rotate(axis_vec3D, angle_degrees, origin_pt3D) for duct_element in self.exhaust_ducting
        ]
        return new_system

    def rotate_xy(self, angle_degrees, origin_pt3D):
        """Rotate the System's ducts counterclockwise in the XY plane by a certain angle.

        Args:
            angle_degrees: An angle in degrees.
            origin_pt3D: A Point3D for the origin around which the object will be rotated.
        """
        new_system = self.duplicate()
        new_system.identifier = self.identifier
        new_system.supply_ducting = [
            duct_element.rotate_xy(angle_degrees, origin_pt3D) for duct_element in self.supply_ducting
        ]
        new_system.exhaust_ducting = [
            duct_element.rotate_xy(angle_degrees, origin_pt3D) for duct_element in self.exhaust_ducting
        ]
        return new_system

    def reflect(self, normal_vec3D, origin_pt3D):
        """Reflected the System's ducts across a plane with the input normal vector and origin.

        Args:
            normal_vec3D: A Vector3D representing the normal vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin_pt3D: A Point3D representing the origin from which to reflect.
        """
        new_system = self.duplicate()
        new_system.identifier = self.identifier
        new_system.supply_ducting = [
            duct_element.reflect(normal_vec3D, origin_pt3D) for duct_element in self.supply_ducting
        ]
        new_system.exhaust_ducting = [
            duct_element.reflect(normal_vec3D, origin_pt3D) for duct_element in self.exhaust_ducting
        ]
        return new_system

    def scale(self, scale_factor, origin_pt3D=None):
        # type: (float, Optional[Point3D]) -> PhVentilationSystem
        """Scale the System's ducts by a factor from an origin point.

        Args:
            scale_factor: A number representing how much the line segment should be scaled.
            origin_pt3D: A Point3D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0, 0).
        """
        new_system = self.duplicate()
        new_system.identifier = self.identifier
        new_system.supply_ducting = [
            duct_element.scale(scale_factor, origin_pt3D) for duct_element in self.supply_ducting
        ]
        new_system.exhaust_ducting = [
            duct_element.scale(scale_factor, origin_pt3D) for duct_element in self.exhaust_ducting
        ]
        return new_system


# -----------------------------------------------------------------------------
# -- Exhaust Ventilators are not part of the Ventilation System,
# -- but instead are treated more like appliances which get added to the Room.


class _ExhaustVentilatorBase(_base._PhHVACBase):
    def __init__(self):
        super(_ExhaustVentilatorBase, self).__init__()
        self.device_class_name = self.__class__.__name__
        self.display_name = "_unnamed_exhaust_ventilator_"
        self.quantity = 1
        self.annual_runtime_minutes = 0.0
        self.exhaust_flow_rate_m3s = 0.0

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = super(_ExhaustVentilatorBase, self).to_dict()
        d["device_class_name"] = self.device_class_name
        d["quantity"] = self.quantity
        d["annual_runtime_minutes"] = self.annual_runtime_minutes
        d["exhaust_flow_rate_m3s"] = self.exhaust_flow_rate_m3s
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> _ExhaustVentilatorBase
        new_obj = cls()
        new_obj.display_name = _input_dict["display_name"]
        new_obj.identifier = _input_dict["identifier"]
        new_obj.user_data = _input_dict.get("user_data", {})
        new_obj.device_class_name = _input_dict["device_class_name"]
        new_obj.quantity = _input_dict["quantity"]
        new_obj.annual_runtime_minutes = _input_dict["annual_runtime_minutes"]
        new_obj.exhaust_flow_rate_m3s = _input_dict["exhaust_flow_rate_m3s"]
        return new_obj

    def __lt__(self, other):
        # type: (_ExhaustVentilatorBase) -> bool
        return self.identifier < other.identifier

    def __str__(self):
        return "{}(display_name={!r}, exhaust_flow_rate_m3s={:0.02f})".format(
            self.__class__.__name__, self.display_name, self.exhaust_flow_rate_m3s
        )

    def __repr__(self):
        return str(self)

    def ToString(self):
        return self.__repr__()

    def move(self, moving_vec3D):
        """Move the device's elements along a vector.

        Args:
            moving_vec3D: A Vector3D with the direction and distance to move the ray.
        """
        pass

    def rotate(self, axis_vec3D, angle_degrees, origin_pt3D):
        """Rotate the device's elements by a certain angle around an axis_vec3D and origin_pt3D.

        Right hand rule applies:
        If axis_vec3D has a positive orientation, rotation will be clockwise.
        If axis_vec3D has a negative orientation, rotation will be counterclockwise.

        Args:
            axis_vec3D: A Vector3D axis_vec3D representing the axis_vec3D of rotation.
            angle_degrees: An angle for rotation in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        """
        pass

    def rotate_xy(self, angle_degrees, origin_pt3D):
        """Rotate the device's elements counterclockwise in the XY plane by a certain angle.

        Args:
            angle_degrees: An angle in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        """
        pass

    def reflect(self, normal_vec3D, origin_pt3D):
        """Reflected the device's elements across a plane with the input normal vector and origin_pt3D.

        Args:
            normal_vec3D: A Vector3D representing the normal vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin_pt3D: A Point3D representing the origin_pt3D from which to reflect.
        """
        pass

    def scale(self, scale_factor, origin_pt3D=None):
        """Scale the device's elements by a factor from an origin_pt3D point.

        Args:
            scale_factor: A number representing how much the line segment should be scaled.
            origin_pt3D: A Point3D representing the origin_pt3D from which to scale.
                If None, it will be scaled from the World origin_pt3D (0, 0, 0).
        """
        pass


class ExhaustVentDryer(_ExhaustVentilatorBase):
    def __init__(self):
        super(ExhaustVentDryer, self).__init__()
        self.display_name = "_unnamed_dryer_exh_"


class ExhaustVentKitchenHood(_ExhaustVentilatorBase):
    def __init__(self):
        super(ExhaustVentKitchenHood, self).__init__()
        self.display_name = "_unnamed_kitchen_hood_exh_"


class ExhaustVentUserDefined(_ExhaustVentilatorBase):
    def __init__(self):
        super(ExhaustVentUserDefined, self).__init__()
        self.display_name = "_unnamed_user_defined_exh_"


class PhExhaustDeviceBuilder(object):
    """Constructor class for HBPH-Exhaust Ventilation Devices"""

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict[str, Any]) -> _ExhaustVentilatorBase
        """Find the right device constructor class from the module based on the 'type' name."""
        device_class_name = _input_dict["device_class_name"]  # type: str

        valid_class_types = [nm for nm in dir(sys.modules[__name__]) if nm.startswith("ExhaustVent")]

        if device_class_name not in valid_class_types:
            raise UnknownPhExhaustVentTypeError(valid_class_types, device_class_name)
        device_class = getattr(sys.modules[__name__], device_class_name)  # type: _ExhaustVentilatorBase
        new_equipment = device_class.from_dict(_input_dict)
        return new_equipment

    def __str__(self):
        return "{}()".format(self.__class__.__name__)

    def __repr__(self):
        return str(self)

    def ToString(self):
        """Overwrite .NET ToString."""
        return repr(self)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Utility functions for converting Honeybee schedules into WUFI schedules."""

from honeybee import room

from honeybee_ph_utils import histogram, ventilation


class SchedItem:
    def __init__(self, _av, _fr):
        # type: (float, float) -> None
        self.period_speed = _av
        self.period_operating_hours = _fr

    def __str__(self):
        return "{}(average_value={!r}, frequency={!r})".format(
            self.__class__.__name__, self.period_speed, self.period_operating_hours
        )

    def __repr__(self):
        return str(self)


class FourPartSched:
    def __init__(self, _h, _s, _b, _m):
        # type: (SchedItem, SchedItem, SchedItem, SchedItem) -> None
        self.high = _h
        self.standard = _s
        self.basic = _b
        self.minimum = _m

    def __str__(self):
        return "{}({!r}, {!r}, {!r}, {!r})".format(
            self.__class__.__name__, self.high, self.standard, self.basic, self.minimum
        )

    def __repr__(self):
        return str(self)


def calc_four_part_vent_sched_values_from_hb_room(_hb_room, _use_dcv=True):
    # type: (room.Room, bool) -> FourPartSched
    """Returns a WUFI-Style four_part schedule values for the Ventilation airflow, based on the HB Room.

    Arguments:
    ----------
        * _hb_room (): The Honeybee Room to build the schedule for.
        * _use_dcv (bool): Use Demand-Controlled Ventilation? default=True. Set 'True' in
            order to take the Occupancy Schedule and Airflow-per-person loads into account.
            If False, will assume constant airflow for occupancy-related ventilation loads.

    Returns:
    --------
        * namedtuple: ie:
            (
                high=('average_value':18, 'frequency':0.25),
                standard=('average_value':12, 'frequency':0.25),
                basic=('average_value':10, 'frequency':0.25),
                minimum=('average_value':8, 'frequency':0.25),
            )
    """

    # -------------------------------------------------------------------------
    # 1) Calc the Peak Airflow Loads (for Ventilation, for Occupancy)
    vent_m3s_total = ventilation.hb_room_peak_ventilation_airflow_by_zone(_hb_room)
    occ_m3s_total = ventilation.hb_room_peak_ventilation_airflow_by_occupancy(_hb_room)

    # -------------------------------------------------------------------------
    # 2) Get the Occupancy + Ventilation Schedules hourly value generators
    # -- Use try/ excepting since some honeybee programs don't have these attrs.
    # -- If schedule is missing, use a default constant value (1) schedule.
    try:
        schd_occ_values = _hb_room.properties.energy.people.occupancy_schedule.values()
    except AttributeError:
        schd_occ_values = (1 for _ in range(8760))

    try:
        schd_vent_values = _hb_room.properties.energy.ventilation.schedule.values()
    except AttributeError:
        schd_vent_values = (1 for _ in range(8760))

    # -------------------------------------------------------------------------
    # 3) Calc the Hourly Airflows, taking the Schedules into account
    # Note: Should BOTH the by_zone AND the by_person flow-rates be modulated?
    # currently yes: both are being affected. Should ONLY the by_person rates be affected?
    if _use_dcv:
        # -- YES DCV = Modulate the flow rates based on occupancy level.
        hourly_m3s_for_occ = (occ_m3s_total * _ for _ in schd_occ_values)
        hourly_m3s_for_vent = (vent_m3s_total * _ for _ in schd_vent_values)
    else:
        # -- No DCV = Use constant flow rate, regardless of occupancy level.
        hourly_m3s_for_occ = (occ_m3s_total * 1 for _ in schd_occ_values)
        hourly_m3s_for_vent = (vent_m3s_total * 1 for _ in schd_vent_values)

    #  ------------------------------------------------------------------------
    # 4) Calc the Percentage of Peak airflow for each hourly value
    peak_total_m3s = vent_m3s_total + occ_m3s_total
    if peak_total_m3s == 0:
        return FourPartSched(
            SchedItem(1.0, 1.0),
            SchedItem(0.0, 0.0),
            SchedItem(0.0, 0.0),
            SchedItem(0.0, 0.0),
        )

    hourly_total_vent_percentage_rate = [
        (a + b) / peak_total_m3s for a, b in zip(hourly_m3s_for_vent, hourly_m3s_for_occ)
    ]

    #  ------------------------------------------------------------------------
    # 6) Histogram that shit
    # -- Average_value = the speed, 'frequency' = hours per day
    four_part_sched_dict = histogram.generate_histogram(
        _data=hourly_total_vent_percentage_rate,
        _num_bins=4,
    )

    # --- Organize Output
    output = FourPartSched(
        SchedItem(
            four_part_sched_dict.get(0, {}).get("average_value", 0),
            four_part_sched_dict.get(0, {}).get("frequency", 0) * 24,
        ),
        SchedItem(
            four_part_sched_dict.get(1, {}).get("average_value", 0),
            four_part_sched_dict.get(1, {}).get("frequency", 0) * 24,
        ),
        SchedItem(
            four_part_sched_dict.get(2, {}).get("average_value", 0),
            four_part_sched_dict.get(2, {}).get("frequency", 0) * 24,
        ),
        SchedItem(
            four_part_sched_dict.get(3, {}).get("average_value", 0),
            four_part_sched_dict.get(3, {}).get("frequency", 0) * 24,
        ),
    )

    return output



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-


"""DEPRECATED in favor of the new 'honeybee_phhvac' module."""


from honeybee_energy_ph.hvac import _base


class PhHeatingSystem(_base._PhHVACBase):
    message = "PhHeatingSystem is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatingDirectElectric(PhHeatingSystem):
    message = "PhHeatingDirectElectric is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatingFossilBoiler(PhHeatingSystem):
    message = "PhHeatingFossilBoiler is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatingWoodBoiler(PhHeatingSystem):
    message = "PhHeatingWoodBoiler is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatingDistrict(PhHeatingSystem):
    message = "PhHeatingDistrict is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatingSystemBuilder(object):
    message = "PhHeatingSystemBuilder is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""HB-Room Passive House (PH) Properties."""

try:
    from typing import Any, Dict, List, Optional, Union
except ImportError:
    pass  # Python2.7

try:
    from ladybug_geometry import geometry3d
except ImportError as e:
    raise ImportError("\nFailed to import ladybug_geometry:\n\t{}".format(e))

try:
    from honeybee import room
except ImportError as e:
    raise ImportError("\nFailed to import honeybee:\n\t{}".format(e))

try:
    from honeybee_ph import space
    from honeybee_ph.bldg_segment import BldgSegment
    from honeybee_ph.foundations import PhFoundation, PhFoundationFactory
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_ph:\n\t{}".format(e))

try:
    from honeybee_ph_utils import enumerables
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_ph_utils:\n\t{}".format(e))

"""
room.Room
    ├─ properties: RoomProperties
        ├─ energy: RoomEnergyProperties
        ├─ ph: RoomPhProperties
            ├─ spaces: List[space.Space]    
                ├─ properties: SpaceProperties
                    ├─ ph: SpacePhProperties
                    ├─ energy: SpaceEnergyProperties
                    ├─...
        ├─ ph_hvac: RoomPhHvacProperties
        ├─ ...
"""

# -----------------------------------------------------------------------------


class PhSpecificHeatCapacity(enumerables.CustomEnum):
    allowed = [
        "1-LIGHTWEIGHT",
        "2-MIXED",
        "3-MASSIVE",
    ]

    def __init__(self, _value=1):
        # type: (Union[str, int]) -> None
        super(PhSpecificHeatCapacity, self).__init__(_value)


# -----------------------------------------------------------------------------


class RoomPhProperties(object):
    def __init__(self, _host):
        # type: (Optional[room.Room]) -> None
        self._host = _host
        self.id_num = 0
        self._spaces = list()  # type: List[space.Space]
        self.ph_bldg_segment = BldgSegment()
        self._ph_foundations = {}  # type: Dict[str, PhFoundation]
        self.specific_heat_capacity = PhSpecificHeatCapacity("1-LIGHTWEIGHT")

    @property
    def spaces(self):
        # type: () -> List[space.Space]
        return self._spaces

    @property
    def total_space_floor_area(self):
        # type: () -> float
        """The total unweighted floor-area of all spaces hosted by the honeybee-Room."""
        return sum((sp.floor_area for sp in self.spaces))

    @property
    def host(self):
        return self._host

    @property
    def ph_foundations(self):
        # type: () -> List[PhFoundation]
        return list(self._ph_foundations.values())

    def duplicate(self, new_host=None, include_spaces=True):
        # type: (Any, bool) -> RoomPhProperties
        _host = new_host or self._host
        new_obj = RoomPhProperties(_host)
        new_obj.id_num = self.id_num
        new_obj.specific_heat_capacity = PhSpecificHeatCapacity(self.specific_heat_capacity.value)

        if include_spaces:
            for sp in self._spaces:
                new_obj._spaces.append(sp.duplicate(_host))

        new_obj.ph_bldg_segment = self.ph_bldg_segment.duplicate()

        for f in self.ph_foundations:
            new_obj.add_foundation(f.duplicate())

        return new_obj

    def ToString(self):
        return self.__repr__()

    def __repr__(self):
        return "{}: [host: {}]".format(self.__class__.__name__, self.host)

    def to_dict(self, abridged=False):
        # type: (bool) -> Dict[str, Any]
        d = {}

        d["spaces"] = [sp.to_dict() for sp in self.spaces]
        d["specific_heat_capacity"] = self.specific_heat_capacity.value

        if abridged == False:
            d["type"] = "RoomPhProperties"
            d["id_num"] = self.id_num
            d["ph_bldg_segment"] = self.ph_bldg_segment.to_dict()
            d["ph_foundations"] = [f.to_dict() for f in self.ph_foundations]
        else:
            d["type"] = "RoomPhPropertiesAbridged"
            d["ph_bldg_segment_id"] = self.ph_bldg_segment.identifier
            d["ph_foundations"] = [f.to_dict() for f in self.ph_foundations]

        return {"ph": d}

    @classmethod
    def from_dict(cls, _input_dict, host):
        # type: (Dict[str, Any], Any) -> RoomPhProperties
        assert _input_dict["type"] == "RoomPhProperties", "Expected RoomPhProperties. Got {}.".format(
            _input_dict["type"]
        )

        new_prop = cls(host)
        new_prop.id_num = _input_dict.get("id_num", 0)
        new_prop.specific_heat_capacity = PhSpecificHeatCapacity(
            _input_dict.get("specific_heat_capacity", new_prop.specific_heat_capacity.value)
        )

        if "ph_bldg_segment" in _input_dict.keys():
            new_prop.ph_bldg_segment = BldgSegment.from_dict(_input_dict.get("ph_bldg_segment", {}))
        else:
            new_prop.ph_bldg_segment = None

        for sp in (space.Space.from_dict(d, host) for d in _input_dict.get("spaces", [])):
            new_prop.add_new_space(sp)

        for f_dict in _input_dict["ph_foundations"]:
            new_prop.add_foundation(PhFoundationFactory.from_dict(f_dict))

        return new_prop

    def apply_properties_from_dict(self, room_prop_dict, bldg_segments):
        # type: (Dict[str, Any], Dict[str, BldgSegment]) -> None
        """Apply properties from a RoomPhPropertiesAbridged dictionary.

        Arguments:
        ----------
            * room_prop_dict (dict): A RoomPhPropertiesAbridged dictionary loaded from
                the room object itself. Unabridged. In Abridged form, this
                dict will just include the 'ph_bldg_segment_id' reference instead of the
                the entire properties data dict.

            * bldg_segments (dict[str: BldgSegment]): A dict of the BldgSegment
                objects found at the Model level. Segment-id is used as the key.

        Returns:
        --------
            * None
        """

        self.specific_heat_capacity = PhSpecificHeatCapacity(
            room_prop_dict.get("specific_heat_capacity", self.specific_heat_capacity.value)
        )

        # -- Set the bldg-segment attributes from the values stored at the 'Model' level
        room_ph_bldg_segment_id = room_prop_dict.get("ph_bldg_segment_id", None)
        if room_ph_bldg_segment_id:
            self.ph_bldg_segment = bldg_segments[room_ph_bldg_segment_id]

        # -- Rebuild the Spaces hosted on the room
        space_dicts = room_prop_dict.get("spaces", [])

        for space_dict in space_dicts:
            self.add_new_space(space.Space.from_dict(space_dict, self.host))

        for f_dict in room_prop_dict["ph_foundations"]:
            self.add_foundation(PhFoundationFactory.from_dict(f_dict))

        return None

    def add_new_space(self, _new_space):
        # type: (space.Space) -> None
        """Adds a new PH-Space to the RoomProperties collection."""
        if _new_space:
            self._spaces.append(_new_space)

    def add_foundation(self, _ph_foundation):
        # type: (PhFoundation) -> None
        if not _ph_foundation:
            return
        self._ph_foundations[_ph_foundation.identifier] = _ph_foundation

    def merge_new_space(self, _new_space):
        # type: (space.Space) -> None
        """Try and merge a new Space with the existing Space.

        If there is no existing Space, will set this as the Room's Space.
        """
        try:
            self._spaces[0].add_new_volumes(_new_space.volumes)
        except IndexError:
            self.add_new_space(_new_space)

    def scale(self, factor, origin=None):
        # type: (float, Optional[geometry3d.Point3D]) -> None
        """Scale the room, and all the spaces in the room by a specified factor.

        Arguments:
        ----------
            * factor (float): The scale factor
            * origin (Optional[geometry3d.Point3D]): default=None, A ladybug_geometry
                Point3D representing the origin from which to scale. If None,
                it will be scaled from the World origin (0, 0, 0).

        Returns:
        --------
            * None
        """

        for space in self.spaces:
            space.scale(factor, origin=None)


def get_ph_prop_from_room(_room):
    # type: (room.Room) -> RoomPhProperties
    """Get the RoomPhProperties of a HB-Room object."""
    return getattr(_room.properties, "ph")



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""PH Foundation Objects."""

from copy import copy

try:
    from typing import Any, Dict, List, Union
except ImportError:
    pass  # Python2.7

try:
    from honeybee_ph import _base
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_ph:\n\t{}".format(e))

try:
    from honeybee_ph_utils import enumerables
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_ph_utils:\n\t{}".format(e))


# -----------------------------------------------------------------------------


class PhFoundationType(enumerables.CustomEnum):
    allowed = [
        "1-HEATED_BASEMENT",
        "2-UNHEATED_BASEMENT",
        "3-SLAB_ON_GRADE",
        "4-VENTED_CRAWLSPACE",
        "5-NONE",
    ]

    def __init__(self, _value=1):
        # type: (Union[str, int]) -> None
        super(PhFoundationType, self).__init__(_value)


class PhSlabEdgeInsulationPosition(enumerables.CustomEnum):
    allowed = [
        "1-UNDEFINED",
        "2-HORIZONTAL",
        "3-VERTICAL",
    ]

    def __init__(self, _value=1):
        # type: (Union[str, int]) -> None
        super(PhSlabEdgeInsulationPosition, self).__init__(_value)


# -----------------------------------------------------------------------------
# -- Base Class


class PhFoundation(_base._Base):
    def __init__(self):
        super(PhFoundation, self).__init__()
        self.foundation_type = PhFoundationType("5-NONE")

    def duplicate(self):
        # type: () -> PhFoundation
        new_obj = self.__class__()
        new_obj.display_name = self.display_name
        new_obj.identifier = self.identifier
        new_obj.user_data = self.user_data
        new_obj.foundation_type = PhFoundationType(self.foundation_type.value)
        new_obj.user_data = self.user_data
        return new_obj

    def __copy__(self):
        return self.duplicate()

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = {}

        d["identifier"] = str(self.identifier)
        d["display_name"] = self.display_name
        d["user_data"] = copy(self.user_data)
        d["foundation_type_value"] = self.foundation_type.value
        d["user_data"] = copy(self.user_data)

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhFoundation
        new_obj = cls()

        new_obj.identifier = _input_dict["identifier"]
        new_obj.display_name = _input_dict["display_name"]
        new_obj.user_data = _input_dict.get("user_data", {})
        new_obj.foundation_type = PhFoundationType(_input_dict["foundation_type_value"])
        new_obj.user_data = _input_dict.get("user_data", {})

        return new_obj

    def base_attrs_from_dict(self, _obj, _input_dict):
        # type: (PhFoundation, dict) -> None
        """Set the base object attributes from a dictionary

        Arguments:
        ----------
            * _obj (PhFoundation): The PH Foundation object to set the attributes of.
            * _input_dict (dict): The dictionary to get the attribute values from.

        Returns:
        --------
            * None
        """
        for attr_name in vars(self).keys():
            try:
                # Strip off underscore so it uses the property setters
                if attr_name.startswith("_"):
                    attr_name = attr_name[1:]
                setattr(_obj, attr_name, _input_dict[attr_name])
            except KeyError:
                pass
        return None

    def __str__(self):
        return "{}(display_name: {}, type: {})".format(
            self.__class__.__name__, self.display_name, self.foundation_type.value
        )

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)


# -----------------------------------------------------------------------------
# -- Specific Foundation Types


class PhHeatedBasement(PhFoundation):
    def __init__(self):
        super(PhHeatedBasement, self).__init__()
        self.foundation_type = PhFoundationType("1-HEATED_BASEMENT")
        self.floor_slab_area_m2 = 0.0
        self.floor_slab_u_value = 1.0
        self.floor_slab_exposed_perimeter_m = 0.0
        self.slab_depth_below_grade_m = 2.5
        self.basement_wall_u_value = 1.0

    def duplicate(self):
        # type: () -> PhHeatedBasement
        obj = self.__class__()
        obj.display_name = self.display_name
        obj.identifier = self.identifier
        obj.user_data = self.user_data
        obj.foundation_type = self.foundation_type
        obj.floor_slab_area_m2 = self.floor_slab_area_m2
        obj.floor_slab_u_value = self.floor_slab_u_value
        obj.floor_slab_exposed_perimeter_m = self.floor_slab_exposed_perimeter_m
        obj.slab_depth_below_grade_m = self.slab_depth_below_grade_m
        obj.basement_wall_u_value = self.basement_wall_u_value
        return obj

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = {}
        d.update(super(PhHeatedBasement, self).to_dict())
        d["floor_slab_area_m2"] = self.floor_slab_area_m2
        d["floor_slab_u_value"] = self.floor_slab_u_value
        d["floor_slab_exposed_perimeter_m"] = self.floor_slab_exposed_perimeter_m
        d["slab_depth_below_grade_m"] = self.slab_depth_below_grade_m
        d["basement_wall_u_value"] = self.basement_wall_u_value
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHeatedBasement
        new_obj = cls()
        super(PhHeatedBasement, new_obj).base_attrs_from_dict(new_obj, _input_dict)
        new_obj.floor_slab_area_m2 = _input_dict["floor_slab_area_m2"]
        new_obj.floor_slab_u_value = _input_dict["floor_slab_u_value"]
        new_obj.floor_slab_exposed_perimeter_m = _input_dict["floor_slab_exposed_perimeter_m"]
        new_obj.slab_depth_below_grade_m = _input_dict["slab_depth_below_grade_m"]
        new_obj.basement_wall_u_value = _input_dict["basement_wall_u_value"]
        return new_obj


class PhUnheatedBasement(PhFoundation):
    def __init__(self):
        super(PhUnheatedBasement, self).__init__()
        self.foundation_type = PhFoundationType("2-UNHEATED_BASEMENT")
        self.floor_ceiling_area_m2 = 0.0
        self.ceiling_u_value = 1.0
        self.floor_slab_exposed_perimeter_m = 0.0
        self.slab_depth_below_grade_m = 0.0
        self.basement_wall_height_above_grade_m = 0.0
        self.basement_wall_uValue_below_grade = 1.0
        self.basement_wall_uValue_above_grade = 1.0
        self.floor_slab_u_value = 1.0
        self.basement_volume_m3 = 0.0
        self.basement_ventilation_ach = 0.0

    def duplicate(self):
        # type: () -> PhUnheatedBasement
        obj = self.__class__()
        obj.display_name = self.display_name
        obj.identifier = self.identifier
        obj.user_data = self.user_data
        obj.foundation_type = self.foundation_type
        obj.floor_ceiling_area_m2 = self.floor_ceiling_area_m2
        obj.ceiling_u_value = self.ceiling_u_value
        obj.floor_slab_exposed_perimeter_m = self.floor_slab_exposed_perimeter_m
        obj.slab_depth_below_grade_m = self.slab_depth_below_grade_m
        obj.basement_wall_height_above_grade_m = self.basement_wall_height_above_grade_m
        obj.basement_wall_uValue_below_grade = self.basement_wall_uValue_below_grade
        obj.basement_wall_uValue_above_grade = self.basement_wall_uValue_above_grade
        obj.floor_slab_u_value = self.floor_slab_u_value
        obj.basement_volume_m3 = self.basement_volume_m3
        obj.basement_ventilation_ach = self.basement_ventilation_ach
        return obj

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = {}
        d.update(super(PhUnheatedBasement, self).to_dict())
        d["floor_ceiling_area_m2"] = self.floor_ceiling_area_m2
        d["ceiling_u_value"] = self.ceiling_u_value
        d["floor_slab_exposed_perimeter_m"] = self.floor_slab_exposed_perimeter_m
        d["slab_depth_below_grade_m"] = self.slab_depth_below_grade_m
        d["basement_wall_height_above_grade_m"] = self.basement_wall_height_above_grade_m
        d["basement_wall_uValue_below_grade"] = self.basement_wall_uValue_below_grade
        d["basement_wall_uValue_above_grade"] = self.basement_wall_uValue_above_grade
        d["floor_slab_u_value"] = self.floor_slab_u_value
        d["basement_volume_m3"] = self.basement_volume_m3
        d["basement_ventilation_ach"] = self.basement_ventilation_ach
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhUnheatedBasement
        new_obj = cls()
        super(PhUnheatedBasement, new_obj).base_attrs_from_dict(new_obj, _input_dict)
        new_obj.floor_ceiling_area_m2 = _input_dict["floor_ceiling_area_m2"]
        new_obj.ceiling_u_value = _input_dict["ceiling_u_value"]
        new_obj.floor_slab_exposed_perimeter_m = _input_dict["floor_slab_exposed_perimeter_m"]
        new_obj.slab_depth_below_grade_m = _input_dict["slab_depth_below_grade_m"]
        new_obj.basement_wall_height_above_grade_m = _input_dict["basement_wall_height_above_grade_m"]
        new_obj.basement_wall_uValue_below_grade = _input_dict["basement_wall_uValue_below_grade"]
        new_obj.basement_wall_uValue_above_grade = _input_dict["basement_wall_uValue_above_grade"]
        new_obj.floor_slab_u_value = _input_dict["floor_slab_u_value"]
        new_obj.basement_volume_m3 = _input_dict["basement_volume_m3"]
        new_obj.basement_ventilation_ach = _input_dict["basement_ventilation_ach"]
        return new_obj


class PhSlabOnGrade(PhFoundation):
    def __init__(self):
        super(PhSlabOnGrade, self).__init__()
        self.foundation_type = PhFoundationType("3-SLAB_ON_GRADE")
        self.floor_slab_area_m2 = 0.0
        self.floor_slab_u_value = 1.0
        self.floor_slab_exposed_perimeter_m = 0.0
        self._perim_insulation_position = PhSlabEdgeInsulationPosition("3-VERTICAL")
        self.perim_insulation_width_or_depth_m = 0.300
        self.perim_insulation_thickness_m = 0.050
        self.perim_insulation_conductivity = 0.04

    @property
    def perim_insulation_position(self):
        return self._perim_insulation_position

    @perim_insulation_position.setter
    def perim_insulation_position(self, _input):
        self._perim_insulation_position = PhSlabEdgeInsulationPosition(_input)

    def duplicate(self):
        # type: () -> PhSlabOnGrade
        obj = self.__class__()
        obj.display_name = self.display_name
        obj.identifier = self.identifier
        obj.user_data = self.user_data
        obj.foundation_type = self.foundation_type
        obj.floor_slab_area_m2 = self.floor_slab_area_m2
        obj.floor_slab_u_value = self.floor_slab_u_value
        obj.floor_slab_exposed_perimeter_m = self.floor_slab_exposed_perimeter_m
        obj._perim_insulation_position = self._perim_insulation_position
        obj.perim_insulation_width_or_depth_m = self.perim_insulation_width_or_depth_m
        obj.perim_insulation_thickness_m = self.perim_insulation_thickness_m
        obj.perim_insulation_conductivity = self.perim_insulation_conductivity
        return obj

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = {}
        d.update(super(PhSlabOnGrade, self).to_dict())
        d["floor_slab_area_m2"] = self.floor_slab_area_m2
        d["floor_slab_u_value"] = self.floor_slab_u_value
        d["floor_slab_exposed_perimeter_m"] = self.floor_slab_exposed_perimeter_m
        d["perim_insulation_position_value"] = self.perim_insulation_position.value
        d["perim_insulation_width_or_depth_m"] = self.perim_insulation_width_or_depth_m
        d["perim_insulation_thickness_m"] = self.perim_insulation_thickness_m
        d["perim_insulation_conductivity"] = self.perim_insulation_conductivity
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhSlabOnGrade
        new_obj = cls()
        super(PhSlabOnGrade, new_obj).base_attrs_from_dict(new_obj, _input_dict)
        new_obj.floor_slab_area_m2 = _input_dict["floor_slab_area_m2"]
        new_obj.floor_slab_u_value = _input_dict["floor_slab_u_value"]
        new_obj.floor_slab_exposed_perimeter_m = _input_dict["floor_slab_exposed_perimeter_m"]
        new_obj.perim_insulation_position = _input_dict["perim_insulation_position_value"]
        new_obj.perim_insulation_width_or_depth_m = _input_dict["perim_insulation_width_or_depth_m"]
        new_obj.perim_insulation_thickness_m = _input_dict["perim_insulation_thickness_m"]
        new_obj.perim_insulation_conductivity = _input_dict["perim_insulation_conductivity"]
        return new_obj


class PhVentedCrawlspace(PhFoundation):
    def __init__(self):
        super(PhVentedCrawlspace, self).__init__()
        self.foundation_type = PhFoundationType("4-VENTED_CRAWLSPACE")
        self.crawlspace_floor_slab_area_m2 = 0.0
        self.ceiling_above_crawlspace_u_value = 1.0
        self.crawlspace_floor_exposed_perimeter_m = 2.5
        self.crawlspace_wall_height_above_grade_m = 0.0
        self.crawlspace_floor_u_value = 1.0
        self.crawlspace_vent_opening_are_m2 = 0.0
        self.crawlspace_wall_u_value = 1.0

    def duplicate(self):
        # type: () -> PhVentedCrawlspace
        obj = self.__class__()
        obj.display_name = self.display_name
        obj.identifier = self.identifier
        obj.user_data = self.user_data
        obj.foundation_type = self.foundation_type
        obj.crawlspace_floor_slab_area_m2 = self.crawlspace_floor_slab_area_m2
        obj.ceiling_above_crawlspace_u_value = self.ceiling_above_crawlspace_u_value
        obj.crawlspace_floor_exposed_perimeter_m = self.crawlspace_floor_exposed_perimeter_m
        obj.crawlspace_wall_height_above_grade_m = self.crawlspace_wall_height_above_grade_m
        obj.crawlspace_floor_u_value = self.crawlspace_floor_u_value
        obj.crawlspace_vent_opening_are_m2 = self.crawlspace_vent_opening_are_m2
        obj.crawlspace_wall_u_value = self.crawlspace_wall_u_value
        return obj

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = {}
        d.update(super(PhVentedCrawlspace, self).to_dict())
        d["crawlspace_floor_slab_area_m2"] = self.crawlspace_floor_slab_area_m2
        d["ceiling_above_crawlspace_u_value"] = self.ceiling_above_crawlspace_u_value
        d["crawlspace_floor_exposed_perimeter_m"] = self.crawlspace_floor_exposed_perimeter_m
        d["crawlspace_wall_height_above_grade_m"] = self.crawlspace_wall_height_above_grade_m
        d["crawlspace_floor_u_value"] = self.crawlspace_floor_u_value
        d["crawlspace_vent_opening_are_m2"] = self.crawlspace_vent_opening_are_m2
        d["crawlspace_wall_u_value"] = self.crawlspace_wall_u_value
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhVentedCrawlspace
        new_obj = cls()
        super(PhVentedCrawlspace, new_obj).base_attrs_from_dict(new_obj, _input_dict)
        new_obj.crawlspace_floor_slab_area_m2 = _input_dict["crawlspace_floor_slab_area_m2"]
        new_obj.ceiling_above_crawlspace_u_value = _input_dict["ceiling_above_crawlspace_u_value"]
        new_obj.crawlspace_floor_exposed_perimeter_m = _input_dict["crawlspace_floor_exposed_perimeter_m"]
        new_obj.crawlspace_wall_height_above_grade_m = _input_dict["crawlspace_wall_height_above_grade_m"]
        new_obj.crawlspace_floor_u_value = _input_dict["crawlspace_floor_u_value"]
        new_obj.crawlspace_vent_opening_are_m2 = _input_dict["crawlspace_vent_opening_are_m2"]
        new_obj.crawlspace_wall_u_value = _input_dict["crawlspace_wall_u_value"]
        return new_obj


# -----------------------------------------------------------------------------
# --- Factory


class PhFoundationFactory(object):
    """Factory class to build any PhFoundation from an input dictionary."""

    type_map = {
        # enum-value: foundation class,
        "1-HEATED_BASEMENT": PhHeatedBasement,
        "2-UNHEATED_BASEMENT": PhUnheatedBasement,
        "3-SLAB_ON_GRADE": PhSlabOnGrade,
        "4-VENTED_CRAWLSPACE": PhVentedCrawlspace,
        "5-NONE": PhFoundation,
    }

    @classmethod
    def _check_input_type_name(cls, _input_type_name):
        # type: (str) -> None
        valid_class_type_names = list(cls.type_map.keys())

        if _input_type_name not in valid_class_type_names:
            msg = 'Error: Unknown PH-Foundation type? Got type of: "{}" but only types: {} are allowed?'.format(
                _input_type_name, valid_class_type_names
            )
            raise Exception(msg)

    @classmethod
    def _get_input_type_name(cls, _input_dict):
        # type: (Dict[str, Any]) -> str
        input_type_enum = PhFoundationType(_input_dict["foundation_type_value"])
        cls._check_input_type_name(input_type_enum.value)
        return input_type_enum.value

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhFoundation

        input_type_name = cls._get_input_type_name(_input_dict)
        foundation_class = cls.type_map[input_type_name]
        new_foundation = foundation_class.from_dict(_input_dict)

        return new_foundation



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Honeybee-PH-HVAC-Equipment: Renewable Energy Devices."""

import sys

try:
    from typing import Any, Dict
except ImportError:
    pass  # IronPython 2.7

try:
    from honeybee_phhvac import _base
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_phhvac:\n\t{}".format(e))


class UnknownPhRenewableEnergyTypeError(Exception):
    def __init__(self, _heater_types, _received_type):
        # type: (list[str], str) -> None
        self.msg = 'Error: Unknown HBPH-Heating-SubSystem type? Got: "{}" but only types: {} are allowed?'.format(
            _received_type, _heater_types
        )
        super(UnknownPhRenewableEnergyTypeError, self).__init__(self.msg)


class PhRenewableEnergyDevice(_base._PhHVACBase):
    """Base class for all HBPH Renewable Energy Systems (PV, etc)."""

    def __init__(self):
        super(PhRenewableEnergyDevice, self).__init__()
        self.device_typename = self.__class__.__name__
        self.percent_coverage = 1.0

    def to_dict(self):
        # type: () -> dict
        d = super(PhRenewableEnergyDevice, self).to_dict()
        d["device_typename"] = self.device_typename
        d["percent_coverage"] = self.percent_coverage
        return d

    def base_attrs_from_dict(self, _input_dict):
        # type: (PhRenewableEnergyDevice, dict) -> PhRenewableEnergyDevice
        self.identifier = _input_dict["identifier"]
        self.display_name = _input_dict["display_name"]
        self.user_data = _input_dict["user_data"]
        self.device_typename = _input_dict["device_typename"]
        self.percent_coverage = _input_dict["percent_coverage"]
        return self

    def check_dict_type(self, _input_dict):
        # type: (dict) -> None
        """Check that the input dict type is correct for the Heating System being constructed."""
        device_type = _input_dict["device_typename"]
        msg = "Error creating Heating System from dict. Expected '{}' but got '{}'".format(
            self.__class__.__name__, device_type
        )
        assert device_type == str(self.__class__.__name__), msg
        return None

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict) -> PhRenewableEnergyDevice
        raise NotImplementedError("Error: from_dict() called on BaseClass.")

    def __lt__(self, other):
        # type: (PhRenewableEnergyDevice) -> bool
        return self.identifier < other.identifier

    def move(self, moving_vec3D):
        """Move the device's elements along a vector.

        Args:
            moving_vec3D: A Vector3D with the direction and distance to move the ray.
        """
        pass

    def rotate(self, axis_vec3D, angle_degrees, origin_pt3D):
        """Rotate the device's elements by a certain angle around an axis_vec3D and origin_pt3D.

        Right hand rule applies:
        If axis_vec3D has a positive orientation, rotation will be clockwise.
        If axis_vec3D has a negative orientation, rotation will be counterclockwise.

        Args:
            axis_vec3D: A Vector3D axis_vec3D representing the axis_vec3D of rotation.
            angle_degrees: An angle for rotation in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        """
        pass

    def rotate_xy(self, angle_degrees, origin_pt3D):
        """Rotate the device's elements counterclockwise in the XY plane by a certain angle.

        Args:
            angle_degrees: An angle in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        """
        pass

    def reflect(self, normal_vec3D, origin_pt3D):
        """Reflected the device's elements across a plane with the input normal vector and origin_pt3D.

        Args:
            normal_vec3D: A Vector3D representing the normal vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin_pt3D: A Point3D representing the origin_pt3D from which to reflect.
        """
        pass

    def scale(self, scale_factor, origin_pt3D=None):
        """Scale the device's elements by a factor from an origin_pt3D point.

        Args:
            scale_factor: A number representing how much the line segment should be scaled.
            origin_pt3D: A Point3D representing the origin_pt3D from which to scale.
                If None, it will be scaled from the World origin_pt3D (0, 0, 0).
        """
        pass


# -----------------------------------------------------------------------------
# Renewable Energy Device Types


class PhPhotovoltaicDevice(PhRenewableEnergyDevice):
    """PV System."""

    def __init__(self):
        super(PhPhotovoltaicDevice, self).__init__()
        self.photovoltaic_renewable_energy = 0.0
        self.array_size = 0.0
        self.utilization_factor = 1.0

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = super(PhPhotovoltaicDevice, self).to_dict()
        d["photovoltaic_renewable_energy"] = self.photovoltaic_renewable_energy
        d["array_size"] = self.array_size
        d["utilization_factor"] = self.utilization_factor
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhPhotovoltaicDevice
        new_obj = cls()
        new_obj.check_dict_type(_input_dict)
        new_obj.base_attrs_from_dict(_input_dict)

        new_obj.photovoltaic_renewable_energy = _input_dict["photovoltaic_renewable_energy"]
        new_obj.array_size = _input_dict["array_size"]
        new_obj.utilization_factor = _input_dict["utilization_factor"]

        return new_obj


# -----------------------------------------------------------------------------


class PhRenewableEnergyDeviceBuilder(object):
    """Constructor class for PH-Renewable-Energy-System objects."""

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict[str, Any]) -> PhRenewableEnergyDevice
        """Find the right device constructor class from the module based on the device_typename."""
        valid_device_type_names = [nm for nm in dir(sys.modules[__name__]) if nm.startswith("Ph")]

        device_typename = _input_dict["device_typename"]
        if device_typename not in valid_device_type_names:
            raise UnknownPhRenewableEnergyTypeError(valid_device_type_names, device_typename)
        device_class = getattr(sys.modules[__name__], device_typename)  # type: PhRenewableEnergyDevice
        new_device = device_class.from_dict(_input_dict)
        return new_device

    def __str__(self):
        return "{}()".format(self.__class__.__name__)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)



# -*- Python Version: 2.7 -*-
# -*- coding: utf-8 -*-

"""HBPH Thermal Bridge Objects"""

try:
    from typing import Any, Union
except ImportError:
    pass  # IronPython 2.7

from ladybug_geometry.geometry3d.polyline import LineSegment3D, Polyline3D

from honeybee_energy_ph.construction import _base
from honeybee_ph_utils import enumerables


class PhThermalBridgeType(enumerables.CustomEnum):
    allowed = [
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        "15-Ambient",
        "16-Perimeter",
        "17-FS/BC",
    ]

    def __init__(self, _value=15, _index_offset=0):
        super(PhThermalBridgeType, self).__init__(_value, _index_offset)


class PhThermalBridge(_base._Base):
    """A single PhThermalBridge object"""

    def __init__(self, _identifier, _geometry):
        # type: (Any, Union[Polyline3D, LineSegment3D]) -> None
        super(PhThermalBridge, self).__init__(_identifier)
        self.geometry = _geometry
        self.display_name = "_unnamed_thermal_bridge_"
        self.quantity = 1.0
        self._group_type = PhThermalBridgeType(15)
        self.psi_value = 0.1
        self.fRsi_value = 0.75

    @property
    def length(self):
        # type: () -> float
        return self.geometry.length

    @property
    def group_type(self):
        # type: () -> PhThermalBridgeType
        return self._group_type

    @group_type.setter
    def group_type(self, _in):
        # type: (Union[str, int]) -> None
        self._group_type = PhThermalBridgeType(_in)

    def to_dict(self):
        # type: () -> dict[str, Any]
        d = super(PhThermalBridge, self).to_dict()
        d["display_name"] = self.display_name
        d["quantity"] = self.quantity
        d["_group_type"] = self._group_type.to_dict()
        d["psi_value"] = self.psi_value
        d["fRsi_value"] = self.fRsi_value
        d["geometry"] = self.geometry.to_dict()

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict) -> PhThermalBridge

        # -- Geom might be either type
        try:
            geom = Polyline3D.from_dict(_input_dict["geometry"])
        except:
            try:
                geom = LineSegment3D.from_dict(_input_dict["geometry"])
            except:
                raise Exception(
                    "Thermal Bridge geometry dict: '{}' is not LineSegment3D or Polyline3D?".format(_input_dict["type"])
                )

        new_obj = cls(_input_dict["identifier"], geom)
        new_obj.set_base_attrs_from_dict(_input_dict)
        new_obj.display_name = _input_dict["display_name"]
        new_obj.quantity = _input_dict["quantity"]
        new_obj._group_type = PhThermalBridgeType.from_dict(_input_dict["_group_type"])
        new_obj.psi_value = _input_dict["psi_value"]
        new_obj.fRsi_value = _input_dict["fRsi_value"]
        return new_obj

    def duplicate(self):
        # type: () -> PhThermalBridge
        return self.__copy__()

    def __copy__(self):
        # type: () -> PhThermalBridge
        new_obj = self.__class__(self.identifier, self.geometry)
        new_obj.set_base_attrs_from_obj(self)
        new_obj.display_name = self.display_name
        new_obj.quantity = self.quantity
        new_obj.group_type = self.group_type.value
        new_obj.psi_value = self.psi_value
        new_obj.fRsi_value = self.fRsi_value
        return new_obj

    def __str__(self):
        return "{}(geometry={}, length={}, display_name={}, psi_value={:.3f}, fRsi_value={:.3f}, length={:.3f})".format(
            self.__class__.__name__,
            self.geometry,
            self.length,
            self.display_name,
            self.psi_value,
            self.fRsi_value,
            self.length,
        )

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""This is called during __init__ and extends the base honeybee class Properties with a new ._ph slot"""

from honeybee.properties import ApertureProperties, FaceProperties, ModelProperties, RoomProperties, ShadeProperties

from .properties.aperture import AperturePhProperties
from .properties.face import FacePhProperties
from .properties.model import ModelPhProperties
from .properties.room import RoomPhProperties
from .properties.shade import ShadePhProperties
from .properties.space import SpacePhProperties, SpaceProperties

# Step 1)
# set a private ._ph attribute on each relevant HB-Core Property class to None

setattr(ModelProperties, "_ph", None)
setattr(RoomProperties, "_ph", None)
setattr(FaceProperties, "_ph", None)
setattr(ApertureProperties, "_ph", None)
setattr(SpaceProperties, "_ph", None)
setattr(ShadeProperties, "_ph", None)

# Step 2)
# create methods to define the public .ph property instances on each obj.properties container


def model_ph_properties(self):
    if self._ph is None:
        self._ph = ModelPhProperties(self.host)
    return self._ph


def room_ph_properties(self):
    if self._ph is None:
        self._ph = RoomPhProperties(self.host)
    return self._ph


def face_ph_properties(self):
    if self._ph is None:
        self._ph = FacePhProperties(self.host)
    return self._ph


def aperture_ph_properties(self):
    if self._ph is None:
        self._ph = AperturePhProperties(self.host)
    return self._ph


def space_ph_properties(self):
    if self._ph is None:
        self._ph = SpacePhProperties(self.host)
    return self._ph


def shade_ph_properties(self):
    if self._ph is None:
        self._ph = ShadePhProperties(self.host)
    return self._ph


# Step 3)
# add public .ph property methods to the Properties classes
setattr(ModelProperties, "ph", property(model_ph_properties))
setattr(RoomProperties, "ph", property(room_ph_properties))
setattr(FaceProperties, "ph", property(face_ph_properties))
setattr(ApertureProperties, "ph", property(aperture_ph_properties))
setattr(SpaceProperties, "ph", property(space_ph_properties))
setattr(ShadeProperties, "ph", property(shade_ph_properties))



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Honeybee-PH-HVAC-Equipment: Aux. Energy Supportive Devices."""

try:
    from typing import Any, Dict
except ImportError:
    pass  # IronPython 2.7

try:
    from honeybee_phhvac import _base
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_phhvac:\n\t{}".format(e))


class PhSupportiveDevice(_base._PhHVACBase):
    def __init__(self):
        # type: () -> None
        super(PhSupportiveDevice, self).__init__()
        self.display_name = "__unnamed_device__"
        self.device_type = 10
        self.quantity = 1
        self.in_conditioned_space = True
        self.norm_energy_demand_W = 1.0
        self.annual_period_operation_khrs = 8.760  # 100% of the year

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = super(PhSupportiveDevice, self).to_dict()
        d["device_class_name"] = self.__class__.__name__
        d["display_name"] = self.display_name
        d["device_type"] = self.device_type
        d["quantity"] = self.quantity
        d["in_conditioned_space"] = self.in_conditioned_space
        d["norm_energy_demand_W"] = self.norm_energy_demand_W
        d["annual_period_operation_khrs"] = self.annual_period_operation_khrs
        return d

    def base_attrs_from_dict(self, _input_dict):
        # type: (Dict[str, Any]) -> PhSupportiveDevice
        self.identifier = _input_dict["identifier"]
        self.display_name = _input_dict["display_name"]
        self.user_data = _input_dict.get("user_data", {})
        self.device_type = _input_dict["device_type"]
        self.quantity = _input_dict["quantity"]
        self.in_conditioned_space = _input_dict["in_conditioned_space"]
        self.norm_energy_demand_W = _input_dict["norm_energy_demand_W"]
        self.annual_period_operation_khrs = _input_dict["annual_period_operation_khrs"]
        return self

    def check_dict_type(self, _input_dict):
        # type: (Dict[str, Any]) -> None
        """Check that the input dict type is correct for the Supportive Device being constructed."""
        device_class_name = _input_dict["device_class_name"]
        msg = "Error creating Supportive Device from dict. Expected '{}' but got '{}'".format(
            self.__class__.__name__, device_class_name
        )
        assert device_class_name == str(self.__class__.__name__), msg
        return None

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhSupportiveDevice
        new_obj = cls()
        new_obj.check_dict_type(_input_dict)
        new_obj.base_attrs_from_dict(_input_dict)
        return new_obj

    def __lt__(self, other):
        # type: (PhSupportiveDevice) -> bool
        return self.identifier < other.identifier

    def __str__(self):
        return "{}(display_name={!r}, device_type={!r}, quantity={!r})".format(
            self.__class__.__name__, self.display_name, self.device_type, self.quantity
        )

    def __repr__(self):
        return str(self)

    def ToString(self):
        return self.__repr__()

    def move(self, moving_vec3D):
        """Move the device's elements along a vector.

        Args:
            moving_vec3D: A Vector3D with the direction and distance to move the ray.
        """
        pass

    def rotate(self, axis_vec3D, angle_degrees, origin_pt3D):
        """Rotate the device's elements by a certain angle around an axis_vec3D and origin_pt3D.

        Right hand rule applies:
        If axis_vec3D has a positive orientation, rotation will be clockwise.
        If axis_vec3D has a negative orientation, rotation will be counterclockwise.

        Args:
            axis_vec3D: A Vector3D axis_vec3D representing the axis_vec3D of rotation.
            angle_degrees: An angle for rotation in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        """
        pass

    def rotate_xy(self, angle_degrees, origin_pt3D):
        """Rotate the device's elements counterclockwise in the XY plane by a certain angle.

        Args:
            angle_degrees: An angle in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        """
        pass

    def reflect(self, normal_vec3D, origin_pt3D):
        """Reflected the device's elements across a plane with the input normal vector and origin_pt3D.

        Args:
            normal_vec3D: A Vector3D representing the normal vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin_pt3D: A Point3D representing the origin_pt3D from which to reflect.
        """
        pass

    def scale(self, scale_factor, origin_pt3D=None):
        """Scale the device's elements by a factor from an origin_pt3D point.

        Args:
            scale_factor: A number representing how much the line segment should be scaled.
            origin_pt3D: A Point3D representing the origin_pt3D from which to scale.
                If None, it will be scaled from the World origin_pt3D (0, 0, 0).
        """
        pass



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""This is called during __init__ and extends the base honeybee class Properties with a new ._ph slot"""

from honeybee_energy.properties.extension import (
    ElectricEquipmentProperties,
    EnergyMaterialNoMassProperties,
    EnergyMaterialProperties,
    EnergyMaterialVegetationProperties,
    LightingProperties,
    OpaqueConstructionProperties,
    PeopleProperties,
    ServiceHotWaterProperties,
    WindowConstructionProperties,
    WindowConstructionShadeProperties,
)

# -- honeybee_energy properties objects
from honeybee_energy.schedule.ruleset import ScheduleRulesetProperties

from honeybee_energy_ph.properties.construction.opaque import OpaqueConstructionPhProperties
from honeybee_energy_ph.properties.construction.window import WindowConstructionPhProperties
from honeybee_energy_ph.properties.construction.windowshade import WindowConstructionShadePhProperties
from honeybee_energy_ph.properties.hot_water.hw_program import ServiceHotWaterPhProperties
from honeybee_energy_ph.properties.load.equipment import ElectricEquipmentPhProperties
from honeybee_energy_ph.properties.load.lighting import LightingPhProperties
from honeybee_energy_ph.properties.load.people import PeoplePhProperties
from honeybee_energy_ph.properties.materials.opaque import (
    EnergyMaterialNoMassPhProperties,
    EnergyMaterialPhProperties,
    EnergyMaterialVegetationPhProperties,
)

# -- honeybee_energy_ph properties objects
from honeybee_energy_ph.properties.ruleset import ScheduleRulesetPhProperties
from honeybee_energy_ph.properties.space import SpaceEnergyProperties
from honeybee_ph.properties.space import SpaceProperties

# Step 1)
# set a private ._ph attribute on each relevant HB-Energy Property class to None
setattr(SpaceProperties, "_energy", None)
setattr(ScheduleRulesetProperties, "_ph", None)
setattr(OpaqueConstructionProperties, "_ph", None)
setattr(EnergyMaterialProperties, "_ph", None)
setattr(EnergyMaterialNoMassProperties, "_ph", None)
setattr(EnergyMaterialVegetationProperties, "_ph", None)
setattr(WindowConstructionProperties, "_ph", None)
setattr(WindowConstructionShadeProperties, "_ph", None)
setattr(ServiceHotWaterProperties, "_ph", None)
setattr(ElectricEquipmentProperties, "_ph", None)
setattr(PeopleProperties, "_ph", None)
setattr(LightingProperties, "_ph", None)


# Step 2)
# create methods to define the public .property.<extension> @property instances on each obj.properties container


def space_energy_properties(self):
    if self._energy is None:
        self._energy = SpaceEnergyProperties(self.host)
    return self._energy


def schedule_ruleset_ph_properties(self):
    if self._ph is None:
        self._ph = ScheduleRulesetPhProperties(self.host)
    return self._ph


def opaque_construction_ph_properties(self):
    if self._ph is None:
        self._ph = OpaqueConstructionPhProperties()
    return self._ph


def energy_material_ph_properties(self):
    if self._ph is None:
        self._ph = EnergyMaterialPhProperties()
    return self._ph


def energy_material_no_mass_ph_properties(self):
    if self._ph is None:
        self._ph = EnergyMaterialNoMassPhProperties()
    return self._ph


def energy_material_vegetation_ph_properties(self):
    if self._ph is None:
        self._ph = EnergyMaterialVegetationPhProperties()
    return self._ph


def window_construction_ph_properties(self):
    if self._ph is None:
        self._ph = WindowConstructionPhProperties(self.host)
    return self._ph


def window_construction_shade_ph_properties(self):
    if self._ph is None:
        self._ph = WindowConstructionShadePhProperties(self.host)
    return self._ph


def hot_water_program_ph_properties(self):
    if self._ph is None:
        self._ph = ServiceHotWaterPhProperties(self.host)
    return self._ph


def elec_equip_ph_properties(self):
    if self._ph is None:
        self._ph = ElectricEquipmentPhProperties(self.host)
    return self._ph


def people_ph_properties(self):
    if self._ph is None:
        self._ph = PeoplePhProperties(self.host)
    return self._ph


def lighting_ph_properties(self):
    if self._ph is None:
        self._ph = LightingPhProperties(self.host)
    return self._ph


# Step 3)
# add public .energy or .ph @property methods to the appropriate Properties classes
setattr(SpaceProperties, "energy", property(space_energy_properties))
setattr(ScheduleRulesetProperties, "ph", property(schedule_ruleset_ph_properties))
setattr(OpaqueConstructionProperties, "ph", property(opaque_construction_ph_properties))
setattr(WindowConstructionProperties, "ph", property(window_construction_ph_properties))
setattr(
    WindowConstructionShadeProperties,
    "ph",
    property(window_construction_shade_ph_properties),
)
setattr(EnergyMaterialProperties, "ph", property(energy_material_ph_properties))
setattr(EnergyMaterialNoMassProperties, "ph", property(energy_material_no_mass_ph_properties))
setattr(
    EnergyMaterialVegetationProperties,
    "ph",
    property(energy_material_vegetation_ph_properties),
)
setattr(ServiceHotWaterProperties, "ph", property(hot_water_program_ph_properties))
setattr(ElectricEquipmentProperties, "ph", property(elec_equip_ph_properties))
setattr(PeopleProperties, "ph", property(people_ph_properties))
setattr(LightingProperties, "ph", property(lighting_ph_properties))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setuptools.setup(
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""HB-Room Passive-House-HVAC-Equipment Properties."""

try:
    from typing import Any, Dict, Optional, Set
except ImportError:
    pass  # Python2.7

try:
    from ladybug_geometry import geometry3d
except ImportError as e:
    raise ImportError("\nFailed to import ladybug_geometry:\n\t{}".format(e))

try:
    from honeybee import room
    from honeybee.properties import RoomProperties
except ImportError as e:
    raise ImportError("\nFailed to import honeybee:\n\t{}".format(e))

try:
    from honeybee_ph import space
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_ph:\n\t{}".format(e))

try:
    from honeybee_phhvac.heat_pumps import PhHeatPumpSystem, PhHeatPumpSystemBuilder
    from honeybee_phhvac.heating import PhHeatingSystem, PhHeatingSystemBuilder
    from honeybee_phhvac.hot_water_system import PhHotWaterSystem
    from honeybee_phhvac.renewable_devices import PhRenewableEnergyDevice, PhRenewableEnergyDeviceBuilder
    from honeybee_phhvac.supportive_device import PhSupportiveDevice
    from honeybee_phhvac.ventilation import PhExhaustDeviceBuilder, PhVentilationSystem, _ExhaustVentilatorBase
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_phhvac:\n\t{}".format(e))


class RoomPhHvacProperties_FromDictError(Exception):
    def __init__(self, _expected_types, _input_type):
        self.msg = 'Error: Expected type of "{}". Got: {}'.format(_expected_types, _input_type)
        super(RoomPhHvacProperties_FromDictError, self).__init__(self.msg)


class RoomPhHvacProperties(object):
    def __init__(self, _host):
        # type: (Optional[RoomProperties]) -> None
        self._host = _host
        self.id_num = 0
        self._ventilation_system = None  # type: Optional[PhVentilationSystem]
        self._heating_systems = set()  # type: set[PhHeatingSystem]
        self._heat_pump_systems = set()  # type: set[PhHeatPumpSystem]
        self._exhaust_vent_devices = set()  # type: set[_ExhaustVentilatorBase]
        self._supportive_devices = set()  # type: set[PhSupportiveDevice]
        self._renewable_devices = set()  # type: set[PhRenewableEnergyDevice]
        self._hot_water_system = None  # type: Optional[PhHotWaterSystem]

    @property
    def host(self):
        # type: () -> Optional[RoomProperties]
        return self._host

    @property
    def ventilation_system(self):
        # type: () -> Optional[PhVentilationSystem]
        return self._ventilation_system

    @property
    def heating_systems(self):
        # type: () -> Set[PhHeatingSystem]
        return self._heating_systems

    @property
    def heat_pump_systems(self):
        # type: () -> Set[PhHeatPumpSystem]
        return self._heat_pump_systems

    @property
    def exhaust_vent_devices(self):
        # type: () -> Set[_ExhaustVentilatorBase]
        return self._exhaust_vent_devices

    @property
    def supportive_devices(self):
        # type: () -> Set[PhSupportiveDevice]
        return self._supportive_devices

    @property
    def renewable_devices(self):
        # type: () -> Set[PhRenewableEnergyDevice]
        return self._renewable_devices

    @property
    def hot_water_system(self):
        # type: () -> Optional[PhHotWaterSystem]
        return self._hot_water_system

    def set_ventilation_system(self, _ventilation_system):
        # type: (Optional[PhVentilationSystem]) -> None
        """Set the Ventilation System serving the Room."""
        self._ventilation_system = _ventilation_system

    def add_heating_system(self, _heating_system):
        # type: (Optional[PhHeatingSystem]) -> None
        """Add a Heating System serving the Room."""
        if _heating_system:
            self._heating_systems.add(_heating_system)

    def add_heat_pump_system(self, _heat_pump_system):
        # type: (Optional[PhHeatPumpSystem]) -> None
        """Add a Heat Pump System serving the Room."""
        if _heat_pump_system:
            self._heat_pump_systems.add(_heat_pump_system)

    def add_exhaust_vent_device(self, _exhaust_vent_device):
        # type: (Optional[_ExhaustVentilatorBase]) -> None
        """Add an Exhaust Vent Device serving the Room."""
        if _exhaust_vent_device:
            self._exhaust_vent_devices.add(_exhaust_vent_device)

    def add_supportive_device(self, _supportive_device):
        # type: (Optional[PhSupportiveDevice]) -> None
        """Add a Supportive Device serving the Room."""
        if _supportive_device:
            self._supportive_devices.add(_supportive_device)

    def add_renewable_device(self, _renewable_device):
        # type: (Optional[PhRenewableEnergyDevice]) -> None
        """Add a Renewable Energy Device serving the Room."""
        if _renewable_device:
            self._renewable_devices.add(_renewable_device)

    def set_hot_water_system(self, _hot_water_system):
        # type: (Optional[PhHotWaterSystem]) -> None
        """Set the Hot Water System serving the Room."""
        self._hot_water_system = _hot_water_system

    def clear_systems(self):
        """Clear all the HVAC Systems from the Room."""
        self._ventilation_system = None
        self._heating_systems.clear()
        self._heat_pump_systems.clear()
        self._exhaust_vent_devices.clear()
        self._supportive_devices.clear()
        self._renewable_devices.clear()
        self._hot_water_system = None

    def to_dict(self, abridged=False):
        # type: (bool) -> Dict[str, Any]
        d = {}

        if abridged:
            d["type"] = "RoomPhHvacPropertiesAbridged"
        else:
            d["type"] = "RoomPhHvacProperties"

        d["id_num"] = self.id_num

        d["ventilation_system"] = self.ventilation_system.to_dict() if self.ventilation_system else None

        d["heating_systems"] = [sys.to_dict() for sys in sorted([_ for _ in self.heating_systems if _ is not None])]

        d["heat_pump_systems"] = [sys.to_dict() for sys in sorted([_ for _ in self.heat_pump_systems if _ is not None])]

        d["exhaust_vent_devices"] = [
            sys.to_dict() for sys in sorted([_ for _ in self.exhaust_vent_devices if _ is not None])
        ]

        d["supportive_devices"] = [
            device.to_dict() for device in sorted([_ for _ in self.supportive_devices if _ is not None])
        ]

        d["renewable_devices"] = [
            device.to_dict() for device in sorted([_ for _ in self.renewable_devices if _ is not None])
        ]

        d["hot_water_system"] = self.hot_water_system.to_dict() if self.hot_water_system else None

        return {"ph_hvac": d}

    @classmethod
    def from_dict(cls, _input_dict, host):
        # type: (Dict[str, Any], Optional[RoomProperties]) -> RoomPhHvacProperties
        valid_types = ("RoomPhHvacProperties", "RoomPhHvacPropertiesAbridged")

        if _input_dict["type"] not in valid_types:
            raise RoomPhHvacProperties_FromDictError(valid_types, _input_dict["type"])

        new_prop = cls(host)
        new_prop.id_num = _input_dict.get("id_num", 0)

        vent_sys_dict = _input_dict.get("ventilation_system")
        if vent_sys_dict:
            new_prop.set_ventilation_system(PhVentilationSystem.from_dict(vent_sys_dict))

        for htg_sys_dict in _input_dict.get("heating_systems", []):
            htg_sys = PhHeatingSystemBuilder.from_dict(htg_sys_dict)
            new_prop.add_heating_system(htg_sys)

        for heat_pump_sys_dict in _input_dict.get("heat_pump_systems", []):
            heat_pump_sys = PhHeatPumpSystemBuilder.from_dict(heat_pump_sys_dict)
            new_prop.add_heat_pump_system(heat_pump_sys)

        for exhaust_vent_device_dict in _input_dict.get("exhaust_vent_devices", []):
            exhaust_device = PhExhaustDeviceBuilder.from_dict(exhaust_vent_device_dict)
            new_prop.add_exhaust_vent_device(exhaust_device)

        for supportive_device_dict in _input_dict.get("supportive_devices", []):
            supportive_device = PhSupportiveDevice.from_dict(supportive_device_dict)
            new_prop.add_supportive_device(supportive_device)

        for renewable_device_dict in _input_dict.get("renewable_devices", []):
            renewable_device = PhRenewableEnergyDeviceBuilder.from_dict(renewable_device_dict)
            new_prop.add_renewable_device(renewable_device)

        hot_water_sys_dict = _input_dict.get("hot_water_system")
        if hot_water_sys_dict:
            new_prop.set_hot_water_system(PhHotWaterSystem.from_dict(hot_water_sys_dict))

        return new_prop

    def apply_properties_from_dict(self, room_prop_dict, mech_systems, *args, **kwargs):
        # type: (Dict[str, Any], Dict[str, dict[str, Any]], list, dict) -> None
        """Apply properties from a RoomPhHvacPropertiesAbridged dictionary."""
        self.id_num = room_prop_dict["id_num"]

        # -- Find the identifiers for each of the room's mechanical systems
        vent_sys_dict = room_prop_dict.get("ventilation_system", {}) or {}
        vent_system_id = vent_sys_dict.get("identifier", None)
        heating_system_ids = [sys.get("identifier") for sys in room_prop_dict.get("heating_systems", [])]
        heat_pump_system_ids = [sys.get("identifier") for sys in room_prop_dict.get("heat_pump_systems", [])]
        exhaust_vent_ids = [sys.get("identifier") for sys in room_prop_dict.get("exhaust_vent_devices", [])]
        supportive_device_ids = [sys.get("identifier") for sys in room_prop_dict.get("supportive_devices", [])]
        renewable_device_ids = [sys.get("identifier") for sys in room_prop_dict.get("renewable_devices", [])]
        hot_water_sys_dict = room_prop_dict.get("hot_water_system", {}) or {}
        hot_water_sys_id = hot_water_sys_dict.get("identifier", None)

        # -- Pull out the actual mechanical systems from the dictionary and apply them to the Properties
        self.set_ventilation_system(mech_systems.get("ventilation_systems", {}).get(vent_system_id, None))

        for sys in (mech_systems.get("heating_systems", {}).get(_id) for _id in heating_system_ids):
            self.add_heating_system(sys)

        for sys in (mech_systems.get("heat_pump_systems", {}).get(_id) for _id in heat_pump_system_ids):
            self.add_heat_pump_system(sys)

        for sys in (mech_systems.get("exhaust_vent_devices", {}).get(_id) for _id in exhaust_vent_ids):
            self.add_exhaust_vent_device(sys)

        for sys in (mech_systems.get("supportive_devices", {}).get(_id) for _id in supportive_device_ids):
            self.add_supportive_device(sys)

        for sys in (mech_systems.get("renewable_devices", {}).get(_id) for _id in renewable_device_ids):
            self.add_renewable_device(sys)

        self.set_hot_water_system(mech_systems.get("hot_water_systems", {}).get(hot_water_sys_id, None))

        return None

    def duplicate(self, new_host=None, include_spaces=True):
        # type: (Optional[RoomProperties], bool) -> RoomPhHvacProperties
        return self.__copy__(new_host)

    def __copy__(self, new_host=None):
        # type: (Optional[RoomProperties]) -> RoomPhHvacProperties
        _host = new_host or self._host
        new_obj = RoomPhHvacProperties(_host)
        new_obj.id_num = self.id_num

        new_obj.set_ventilation_system(self.ventilation_system.duplicate() if self.ventilation_system else None)

        for htg_sys in self.heating_systems:
            new_obj.add_heating_system(htg_sys)

        for heat_pump_sys in self.heat_pump_systems:
            new_obj.add_heat_pump_system(heat_pump_sys)

        for exhaust_device in self.exhaust_vent_devices:
            new_obj.add_exhaust_vent_device(exhaust_device)

        for supportive_device in self.supportive_devices:
            new_obj.add_supportive_device(supportive_device)

        for renewable_device in self.renewable_devices:
            new_obj.add_renewable_device(renewable_device)

        new_obj.set_hot_water_system(self.hot_water_system.duplicate() if self.hot_water_system else None)

        return new_obj

    def move(self, move_vec3D):
        """Move the Room's HVAC Systems along a vector.

        When used in within Honeybee, this method will most often be triggered as part of a
        larger model-transform. For instance, when triggered during a Room's
        `honeybee.properties._Properties` transform.

        When this method is used on its own, it is best practice to duplicate the
        object before applying this transform since it applies the transform to the object
        directly, rather than returning a new object with the transform applied.

        Args:
            move_vec3D: A Vector3D with the direction and distance to move the ray.
        """
        if self._ventilation_system:
            self._ventilation_system.move(move_vec3D)

        for sys in self._heating_systems:
            sys.move(move_vec3D)

        for sys in self._heat_pump_systems:
            sys.move(move_vec3D)

        for sys in self._exhaust_vent_devices:
            sys.move(move_vec3D)

        for sys in self._supportive_devices:
            sys.move(move_vec3D)

        for sys in self._renewable_devices:
            sys.move(move_vec3D)

        if self._hot_water_system:
            self._hot_water_system.move(move_vec3D)

    def rotate(self, axis_vec3D, angle_degrees, origin_pt3D):
        """Rotate the Room's HVAC Systems by a certain angle around an axis and origin.

        When used in within Honeybee, this method will most often be triggered as part of a
        larger model-transform. For instance, when triggered during a Room's
        `honeybee.properties._Properties` transform.

        When this method is used on its own, it is best practice to duplicate the
        object before applying this transform since it applies the transform to the object
        directly, rather than returning a new object with the transform applied.

        Right hand rule applies:
        If axis has a positive orientation, rotation will be clockwise.
        If axis has a negative orientation, rotation will be counterclockwise.

        Args:
            axis_vec3D: A Vector3D axis representing the axis of rotation.
            angle_degrees: An angle for rotation in degrees.
            origin_pt3D: A Point3D for the origin around which the object will be rotated.
        """
        if self._ventilation_system:
            self._ventilation_system.rotate(axis_vec3D, angle_degrees, origin_pt3D)

        for sys in self._heating_systems:
            sys.rotate(axis_vec3D, angle_degrees, origin_pt3D)

        for sys in self._heat_pump_systems:
            sys.rotate(axis_vec3D, angle_degrees, origin_pt3D)

        for sys in self._exhaust_vent_devices:
            sys.rotate(axis_vec3D, angle_degrees, origin_pt3D)

        for sys in self._supportive_devices:
            sys.rotate(axis_vec3D, angle_degrees, origin_pt3D)

        for sys in self._renewable_devices:
            sys.rotate(axis_vec3D, angle_degrees, origin_pt3D)

        if self._hot_water_system:
            self._hot_water_system.rotate(axis_vec3D, angle_degrees, origin_pt3D)

    def rotate_xy(self, angle_degree, origin_pt3D):
        # type: (float, geometry3d.Point3D) -> None
        """Rotate the Room's HVAC Systems counterclockwise in the XY plane by a certain angle.

        When used in within Honeybee, this method will most often be triggered as part of a
        larger model-transform. For instance, when triggered during a Room's
        `honeybee.properties._Properties` transform.

        When this method is used on its own, it is best practice to duplicate the
        object before applying this transform since it applies the transform to the object
        directly, rather than returning a new object with the transform applied.

        Args:
            angle_degree: An angle in degrees.
            origin_pt3D: A Point3D for the origin around which the object will be rotated.
        """

        if self._ventilation_system:
            self._ventilation_system = self._ventilation_system.rotate_xy(angle_degree, origin_pt3D)

        for sys in self._heating_systems:
            sys.rotate_xy(angle_degree, origin_pt3D)

        for sys in self._heat_pump_systems:
            sys.rotate_xy(angle_degree, origin_pt3D)

        for sys in self._exhaust_vent_devices:
            sys.rotate_xy(angle_degree, origin_pt3D)

        for sys in self._supportive_devices:
            sys.rotate_xy(angle_degree, origin_pt3D)

        for sys in self._renewable_devices:
            sys.rotate_xy(angle_degree, origin_pt3D)

        if self._hot_water_system:
            self._hot_water_system = self._hot_water_system.rotate_xy(angle_degree, origin_pt3D)

    def reflect(self, normal_vec3D, origin_pt3D):
        """Reflected the Room's HVAC Systems across a plane with the input normal vector and origin.

        When used in within Honeybee, this method will most often be triggered as part of a
        larger model-transform. For instance, when triggered during a Room's
        `honeybee.properties._Properties` transform.

        When this method is used on its own, it is best practice to duplicate the
        object before applying this transform since it applies the transform to the object
        directly, rather than returning a new object with the transform applied.

        Args:
            normal_vec3D: A Vector3D representing the normal vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin_pt3D: A Point3D representing the origin from which to reflect.
        """
        if self._ventilation_system:
            self._ventilation_system.reflect(normal_vec3D, origin_pt3D)

        for sys in self._heating_systems:
            sys.reflect(normal_vec3D, origin_pt3D)

        for sys in self._heat_pump_systems:
            sys.reflect(normal_vec3D, origin_pt3D)

        for sys in self._exhaust_vent_devices:
            sys.reflect(normal_vec3D, origin_pt3D)

        for sys in self._supportive_devices:
            sys.reflect(normal_vec3D, origin_pt3D)

        for sys in self._renewable_devices:
            sys.reflect(normal_vec3D, origin_pt3D)

        if self._hot_water_system:
            self._hot_water_system.reflect(normal_vec3D, origin_pt3D)

    def scale(self, scale_factor, origin_pt3D=None):
        """Scale the Room's HVAC Systems by a factor from an origin point.

        When used in within Honeybee, this method will most often be triggered as part of a
        larger model-transform. For instance, when triggered during a Room's
        `honeybee.properties._Properties` transform.

        When this method is used on its own, it is best practice to duplicate the
        object before applying this transform since it applies the transform to the object
        directly, rather than returning a new object with the transform applied.

        Args:
            scale_factor: A number representing how much the line segment should be scaled.
            origin_pt3D: A Point3D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0, 0).
        """
        if self._ventilation_system:
            self.set_ventilation_system(self._ventilation_system.scale(scale_factor, origin_pt3D))

        for sys in self._heating_systems:
            sys.scale(scale_factor, origin_pt3D)

        for sys in self._heat_pump_systems:
            sys.scale(scale_factor, origin_pt3D)

        for sys in self._exhaust_vent_devices:
            sys.scale(scale_factor, origin_pt3D)

        for sys in self._supportive_devices:
            sys.scale(scale_factor, origin_pt3D)

        for sys in self._renewable_devices:
            sys.scale(scale_factor, origin_pt3D)

        if self._hot_water_system:
            self.set_hot_water_system(self._hot_water_system.scale(scale_factor, origin_pt3D))

    def __str__(self):
        # type: () -> str
        return "{}(host={!r})".format(self.__class__.__name__, self.host)

    def __repr__(self):
        # type: () -> str
        return str(self)

    def ToString(self):
        # type: () -> str
        return self.__repr__()


# -----------------------------------------------------------------------------
# -- Helper Functions to get HVAC Devices and Properties from HB-Room and HB-Space


def get_ph_hvac_prop_from_room(_room):
    # type: (room.Room) -> RoomPhHvacProperties
    """Get the RoomPhHvacProperties of a HB-Room object."""
    return getattr(_room.properties, "ph_hvac")


def get_ph_hvac_from_space(_space):
    # type: (space.Space) -> Optional[RoomPhHvacProperties]
    """Return the "ph_hvac" Properties of a Honeybee-PH Space's host Room."""
    if not _space.host:
        return None
    return getattr(_space.host.properties, "ph_hvac")


def get_ventilation_system_from_space(_space):
    # type: (space.Space) -> Optional[PhVentilationSystem]
    """Get the Ventilation System from a Honeybee-PH Space."""
    ph_hvac = get_ph_hvac_from_space(_space)
    if not ph_hvac:
        return None
    return ph_hvac.ventilation_system


def get_exhaust_vent_devices_from_space(_hph_space):
    # type: (space.Space) -> Set[_ExhaustVentilatorBase]
    """Return a set of all the ExhaustVentilators found on a space's host HB Room."""
    hvac_prop_ph = get_ph_hvac_from_space(_hph_space)
    if not hvac_prop_ph:
        return set()
    return hvac_prop_ph.exhaust_vent_devices


def get_heating_systems_from_space(_space):
    # type: (space.Space)-> Set[PhHeatingSystem]
    """Return the Heating Systems from a Honeybee-PH Space."""
    ph_hvac = get_ph_hvac_from_space(_space)
    if not ph_hvac:
        return set()
    return ph_hvac.heating_systems


def get_heat_pump_systems_from_space(_space):
    # type: (space.Space) -> Set[PhHeatPumpSystem]
    """Return the Heat-Pump Systems from a Honeybee-PH Space."""
    ph_hvac = get_ph_hvac_from_space(_space)
    if not ph_hvac:
        return set()
    return ph_hvac.heat_pump_systems


def get_supportive_devices_from_space(_hph_space):
    # type: ( space.Space) -> Set[PhSupportiveDevice]
    """Return a set of all the SupportiveDevices found on a space's host HB Room."""
    hvac_prop_ph = get_ph_hvac_from_space(_hph_space)
    if not hvac_prop_ph:
        return set()
    else:
        return hvac_prop_ph.supportive_devices


def get_renewable_devices_from_space(_hph_space):
    # type: (space.Space) -> Set[PhRenewableEnergyDevice]
    """Return a set of all the Renewable Energy Devices found on a space's host HB Room."""
    hvac_prop_ph = get_ph_hvac_from_space(_hph_space)
    if not hvac_prop_ph:
        return set()
    else:
        return hvac_prop_ph.renewable_devices


def get_hot_water_system_from_space(_space):
    # type: (space.Space) -> Optional[PhHotWaterSystem]
    """Get the Hot Water System from a Honeybee-PH Space."""
    ph_hvac = get_ph_hvac_from_space(_space)
    if not ph_hvac:
        return None
    return ph_hvac.hot_water_system



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Honeybee-PH-HVAC-Equipment Base Class"""

import uuid
from copy import copy

try:
    from typing import Any, Dict
except ImportError:
    pass  # IronPython 2.7


class _PhHVACBase(object):
    """Base class for any HB-Energy-PH HVAC Objects"""

    def __init__(self):
        self._identifier = uuid.uuid4()
        self.user_data = {}
        self._display_name = self.identifier

    @property
    def identifier(self):
        return str(self._identifier)

    @identifier.setter
    def identifier(self, _in):
        self._identifier = _in

    @property
    def display_name(self):
        """Get or set a string for the object name without any character restrictions.
        If not set, this will be equal to the identifier.
        """
        return self._display_name

    @display_name.setter
    def display_name(self, value):
        try:
            self._display_name = str(value)
        except UnicodeEncodeError:  # Python 2 machine lacking the character set
            self._display_name = value  # keep it as unicode

    @property
    def identifier_short(self):
        # type: () -> str
        return str(self.identifier).split("-")[0]

    @property
    def key(self):
        return self.identifier

    def __hash__(self):
        return hash(self.key)

    def __str__(self):
        return "HB-PH-HVAC_{}: ID-{}".format(self.__class__.__name__, self.identifier_short)

    def ToString(self):
        return str(self)

    def __repr__(self):
        return "{}(identifier={!r}, user_data={!r})".format(
            self.__class__.__name__, self.identifier_short, self.user_data
        )

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = {}
        d["identifier"] = copy(self.identifier)
        d["display_name"] = copy(self.display_name)
        d["user_data"] = copy(self.user_data)
        return d

    def from_dict(self, _input_dict):
        # type: (Dict[str, Any]) -> _PhHVACBase
        obj = self.__class__()
        obj.identifier = _input_dict["identifier"]
        obj.display_name = _input_dict["display_name"]
        obj.user_data = _input_dict["user_data"]
        return obj

    def __copy__(self):
        return self.duplicate()

    def duplicate(self):
        # type: () -> _PhHVACBase
        obj = self.__class__()

        obj.identifier = self.identifier
        obj.display_name = self.display_name
        obj.user_data = copy(self.user_data)

        return obj

    def __eq__(self, other):
        # type: (_PhHVACBase) -> bool
        for k, v in self.__dict__.items():
            try:
                if v != getattr(other, k):
                    if str(v) != str(getattr(other, k)):  # Handle UUID Identifier
                        return False
            except AttributeError:
                return False
        return True
